#! /usr/bin/env python
# encoding: utf-8

import os,types,imp,cPickle,sys,shlex,warnings
from Utils import md5
import Build,Utils,Configure,Task,Options
from Logs import warn,debug
from Constants import*
from Configure import conf,conftest
stdincpath=['/usr/include/','/usr/local/include/']
stdlibpath=['/usr/lib/','/usr/local/lib/','/lib']
class attached_conf(type):
	def __init__(cls,name,bases,dict):
		super(attached_conf,cls).__init__(name,bases,dict)
		def fun_create(self):
			inst=cls(self)
			return inst
		setattr(Configure.ConfigurationContext,'create_'+cls.__name__,fun_create)
class enumerator_base(object):
	def __init__(self,conf):
		self.conf=conf
		self.env=conf.env
		self.define=''
		self.mandatory=0
		self.message=''
	def error(self):
		if not self.message:
			Logs.warn('No message provided')
		self.conf.fatal(self.message)
	def hash(self):
		m=md5()
		classvars=vars(self)
		for(var,value)in classvars.iteritems():
			if callable(var)or value==self or value==self.env or value==self.conf:
				continue
			m.update(str(value))
		return m.digest()
	def update_env(self,hashtable):
		if not type(hashtable)is types.StringType:
			for name in hashtable.keys():
				self.env.append_value(name,hashtable[name])
	def validate(self):
		pass
	def run(self):
		self.validate()
		ret=self.run_test()
		if self.mandatory and not ret:self.error()
		return ret
	def run_test(self):
		return not TEST_OK
class configurator_base(enumerator_base):
	def __init__(self,conf):
		enumerator_base.__init__(self,conf)
		self.uselib_store=''
class function_enumerator(enumerator_base):
	__metaclass__=attached_conf
	def __init__(self,conf):
		enumerator_base.__init__(self,conf)
		self.function=''
		self.define=''
		self.headers=[]
		self.header_code=''
		self.custom_code=''
		self.include_paths=[]
		self.libs=[]
		self.lib_paths=[]
	def error(self):
		errmsg='function %s cannot be found'%self.function
		if self.message:errmsg+='\n%s'%self.message
		self.conf.fatal(errmsg)
	def validate(self):
		if not self.define:
			self.define=self.function.upper()
	def run_test(self):
		ret=not TEST_OK
		oldlibpath=self.env['LIBPATH']
		oldlib=self.env['LIB']
		code=[]
		code.append(self.header_code)
		code.append('\n')
		for header in self.headers:
			code.append('#include <%s>\n'%header)
		if self.custom_code:
			code.append('int main(){%s\nreturn 0;}\n'%self.custom_code)
		else:
			code.append('int main(){\nvoid *p;\np=(void*)(%s);\nreturn 0;\n}\n'%self.function)
		self.env['LIB']=Utils.to_list(self.libs)
		self.env['LIBPATH']=Utils.to_list(self.lib_paths)
		obj=check_data()
		obj.code="\n".join(code)
		obj.includes=self.include_paths
		obj.env=self.env
		ret=int(self.conf.run_check(obj))
		self.conf.check_message('function %s'%self.function,'',ret,option='')
		self.conf.define_cond(self.define,ret)
		self.env['LIB']=oldlib
		self.env['LIBPATH']=oldlibpath
		return ret
class library_enumerator(enumerator_base):
	__metaclass__=attached_conf
	def __init__(self,conf):
		enumerator_base.__init__(self,conf)
		self.name=''
		self.path=[]
		self.code='int main() {return 0;}\n'
		self.uselib_store=''
		self.uselib=''
		self.nosystem=0
		self.want_message=1
	def error(self):
		errmsg='library %s cannot be found'%self.name
		if self.message:errmsg+='\n%s'%self.message
		self.conf.fatal(errmsg)
	def validate(self):
		if not self.nosystem and not self.path:
			self.path+=stdlibpath
	def run_test(self):
		ret=''
		patterns=[self.env['shlib_PATTERN'],'lib%s.dll.a','lib%s.lib',self.env['staticlib_PATTERN']]
		for x in patterns:
			name=x%self.name
			ret=Configure.find_file(name,self.path)
			if ret:break
		if self.want_message:
			self.conf.check_message('library '+self.name,'',ret,option=ret)
		if self.uselib_store:
			self.env.append_value('LIB_'+self.uselib_store,self.name)
			self.env.append_value('LIBPATH_'+self.uselib_store,ret)
		return ret
class header_enumerator(enumerator_base):
	__metaclass__=attached_conf
	def __init__(self,conf):
		enumerator_base.__init__(self,conf)
		self.name=[]
		self.path=[]
		self.define=[]
		self.nosystem=0
		self.want_message=1
	def validate(self):
		if not self.nosystem and not self.path:
			self.path=stdincpath
	def error(self):
		errmsg='cannot find %s in %s'%(self.name,str(self.path))
		if self.message:errmsg+='\n%s'%self.message
		self.conf.fatal(errmsg)
	def run_test(self):
		ret=Configure.find_file(self.name,self.path)
		if self.want_message:
			self.conf.check_message('header',self.name,ret,ret)
		if self.define:self.env[self.define]=ret
		return ret
class cfgtool_configurator(configurator_base):
	__metaclass__=attached_conf
	def __init__(self,conf):
		configurator_base.__init__(self,conf)
		self.uselib_store=''
		self.define=''
		self.binary=''
		self.tests={}
	def error(self):
		errmsg='%s cannot be found'%self.binary
		if self.message:errmsg+='\n%s'%self.message
		self.conf.fatal(errmsg)
	def validate(self):
		if not self.binary:
			raise ValueError,"no binary given in cfgtool!"
		if not self.uselib_store:
			raise ValueError,"no uselib_store given in cfgtool!"
		if not self.define:
			self.define=self.conf.have_define(self.uselib_store)
		if not self.tests:
			self.tests['--cflags']='CCFLAGS'
			self.tests['--cflags']='CXXFLAGS'
			self.tests['--libs']='LINKFLAGS'
	def run_test(self):
		retval={}
		found=TEST_OK
		null='2>/dev/null'
		if sys.platform=="win32":null='2>nul'
		try:
			ret=os.popen('%s %s %s'%(self.binary,self.tests.keys()[0],null)).close()
			if ret:raise ValueError,"error"
			for flag in self.tests:
				var=self.tests[flag]+'_'+self.uselib_store
				cmd='%s %s %s'%(self.binary,flag,null)
				retval[var]=[Utils.cmd_output(cmd).strip()]
			self.update_env(retval)
		except ValueError:
			retval={}
			found=not TEST_OK
		self.conf.define_cond(self.define,found)
		self.conf.check_message('config-tool '+self.binary,'',found,option='')
		return retval
class pkgconfig_configurator(configurator_base):
	__metaclass__=attached_conf
	def __init__(self,conf):
		configurator_base.__init__(self,conf)
		self.name=''
		self.version=''
		self.pkgpath=os.path.join(Options.options.prefix,'lib','pkgconfig')
		self.uselib_store=''
		self.define=''
		self.binary=''
		self.static=False
		self.variables=[]
		self.defines={}
	def error(self):
		if self.version:
			errmsg='pkg-config cannot find %s >= %s'%(self.name,self.version)
		else:
			errmsg='pkg-config cannot find %s'%self.name
		if self.message:errmsg+='\n%s'%self.message
		self.conf.fatal(errmsg)
	def validate(self):
		if not self.uselib_store:
			self.uselib_store=self.name.upper()
		if not self.define:
			self.define=self.conf.have_define(self.uselib_store)
	def _setup_pkg_config_path(self):
		pkgpath=self.pkgpath
		if not pkgpath:
			return""
		if sys.platform=='win32':
			if hasattr(self,'pkgpath_win32_setup'):
				return""
			pkgpath_env=os.getenv('PKG_CONFIG_PATH')
			if pkgpath_env:
				pkgpath_env=pkgpath_env+';'+pkgpath
			else:
				pkgpath_env=pkgpath
			os.putenv('PKG_CONFIG_PATH',pkgpath_env)
			setattr(self,'pkgpath_win32_setup',True)
			return""
		pkgpath='PKG_CONFIG_PATH=$PKG_CONFIG_PATH:'+pkgpath
		return pkgpath
	def run_test(self):
		pkgbin=self.binary
		uselib_store=self.uselib_store
		if type(self.variables)is types.StringType:
			self.variables=str(self.variables).split()
		if not pkgbin:
			pkgbin='pkg-config'
		pkgpath=self._setup_pkg_config_path()
		pkgcom='%s %s'%(pkgpath,pkgbin)
		for key,val in self.defines.items():
			pkgcom+=' --define-variable=%s=%s'%(key,val)
		if self.static:
			pkgcom+=' --static'
		g_defines=self.env['PKG_CONFIG_DEFINES']
		if type(g_defines)is types.DictType:
			for key,val in g_defines.items():
				if self.defines and self.defines.has_key(key):
					continue
				pkgcom+=' --define-variable=%s=%s'%(key,val)
		retval={}
		try:
			if self.version:
				cmd="%s --atleast-version=%s \"%s\""%(pkgcom,self.version,self.name)
				ret=os.popen(cmd).close()
				debug('conf: pkg-config cmd "%s" returned %s'%(cmd,ret))
				self.conf.check_message('package %s >= %s'%(self.name,self.version),'',not ret)
				if ret:raise ValueError,"error"
			else:
				cmd="%s \"%s\""%(pkgcom,self.name)
				ret=os.popen(cmd).close()
				debug('conf: pkg-config cmd "%s" returned %s'%(cmd,ret))
				self.conf.check_message('package %s'%(self.name),'',not ret)
				if ret:
					raise ValueError,"error"
			cflags_I=shlex.split(Utils.cmd_output('%s --cflags-only-I \"%s\"'%(pkgcom,self.name)))
			cflags_other=shlex.split(Utils.cmd_output('%s --cflags-only-other \"%s\"'%(pkgcom,self.name)))
			retval['CCFLAGS_'+uselib_store]=cflags_other
			retval['CXXFLAGS_'+uselib_store]=cflags_other
			retval['CPPPATH_'+uselib_store]=[]
			for incpath in cflags_I:
				assert incpath[:2]=='-I'or incpath[:2]=='/I'
				retval['CPPPATH_'+uselib_store].append(incpath[2:])
			static_l=''
			if self.static:
				static_l='STATIC'
			modlibs=Utils.cmd_output('%s --libs-only-l \"%s\"'%(pkgcom,self.name)).strip().split()
			retval[static_l+'LIB_'+uselib_store]=[]
			for item in modlibs:
				retval[static_l+'LIB_'+uselib_store].append(item[2:])
			modpaths=Utils.cmd_output('%s --libs-only-L \"%s\"'%(pkgcom,self.name)).split()
			retval['LIBPATH_'+uselib_store]=[]
			for item in modpaths:
				retval['LIBPATH_'+uselib_store].append(item[2:])
			modother=Utils.cmd_output('%s --libs-only-other \"%s\"'%(pkgcom,self.name)).strip().split()
			retval['LINKFLAGS_'+uselib_store]=[]
			for item in modother:
				if str(item).endswith(".la"):
					import libtool
					la_config=libtool.libtool_config(item)
					libs_only_L=la_config.get_libs_only_L()
					libs_only_l=la_config.get_libs_only_l()
					for entry in libs_only_l:
						retval[static_l+'LIB_'+uselib_store].append(entry[2:])
					for entry in libs_only_L:
						retval['LIBPATH_'+uselib_store].append(entry[2:])
				else:
					retval['LINKFLAGS_'+uselib_store].append(item)
			for variable in self.variables:
				var_defname=''
				if(type(variable)is types.ListType):
					if len(variable)==2 and variable[1]:
						var_defname=variable[1]
					variable=variable[0]
				if not var_defname:
					var_defname=uselib_store+'_'+variable.upper()
				retval[var_defname]=Utils.cmd_output('%s --variable=%s \"%s\"'%(pkgcom,variable,self.name)).strip()
			self.conf.define(self.define,1)
			self.update_env(retval)
		except ValueError:
			retval={}
			self.conf.undefine(self.define)
		return retval
class test_configurator(configurator_base):
	__metaclass__=attached_conf
	def __init__(self,conf):
		configurator_base.__init__(self,conf)
		self.name=''
		self.code=''
		self.flags=''
		self.define=''
		self.uselib=''
		self.want_message=0
	def error(self):
		errmsg='test program would not run'
		if self.message:errmsg+='\n%s'%self.message
		self.conf.fatal(errmsg)
	def validate(self):
		if not self.code:
			raise Configure.ConfigurationError('test configurator needs code to compile and run!')
	def run_test(self):
		obj=check_data()
		obj.code=self.code
		obj.env=self.env
		obj.uselib=self.uselib
		obj.flags=self.flags
		obj.force_compiler=getattr(self,'force_compiler',None)
		obj.execute=1
		ret=self.conf.run_check(obj)
		if self.want_message:
			if ret:data=ret['result']
			else:data=''
			self.conf.check_message('custom code','',ret,option=data)
		return ret
class library_configurator(configurator_base):
	__metaclass__=attached_conf
	def __init__(self,conf):
		configurator_base.__init__(self,conf)
		self.name=''
		self.path=[]
		self.define=''
		self.nosystem=0
		self.uselib=''
		self.uselib_store=''
		self.static=False
		self.libs=[]
		self.lib_paths=[]
		self.code='int main(){return 0;}\n'
	def error(self):
		errmsg='library %s cannot be linked'%self.name
		if self.message:errmsg+='\n%s'%self.message
		self.conf.fatal(errmsg)
	def validate(self):
		if not self.uselib_store:
			self.uselib_store=self.name.upper()
		if not self.define:
			self.define=self.conf.have_define(self.uselib_store)
		if not self.uselib_store:
			self.conf.fatal('uselib_store is not defined')
		if not self.code:
			self.conf.fatal('library enumerator must have code to compile')
	def run_test(self):
		oldlibpath=self.env['LIBPATH']
		oldlib=self.env['LIB']
		static_l=''
		if self.static:
			static_l='STATIC'
		olduselibpath=list(self.env['LIBPATH_'+self.uselib_store])
		olduselib=list(self.env[static_l+'LIB_'+self.uselib_store])
		test=self.conf.create_library_enumerator()
		test.nosystem=self.nosystem
		test.name=self.name
		test.want_message=0
		test.path=self.path
		test.env=self.env
		ret=test.run()
		if ret:
			self.env['LIBPATH_'+self.uselib_store]+=[ret]
		self.env[static_l+'LIB_'+self.uselib_store]+=[self.name]
		self.env['LIB']=[self.name]+self.libs
		self.env['LIBPATH']=self.lib_paths
		obj=check_data()
		obj.code=self.code
		obj.env=self.env
		obj.uselib=self.uselib_store+" "+self.uselib
		obj.libpath=self.path
		ret=int(self.conf.run_check(obj))
		self.conf.check_message('library %s'%self.name,'',ret)
		self.conf.define_cond(self.define,ret)
		val={}
		if ret:
			val['LIBPATH_'+self.uselib_store]=self.env['LIBPATH_'+self.uselib]
			val[static_l+'LIB_'+self.uselib_store]=self.env['LIB_'+self.uselib]
			val[self.define]=ret
		else:
			self.env['LIBPATH_'+self.uselib_store]=olduselibpath
			self.env[static_l+'LIB_'+self.uselib_store]=olduselib
		self.env['LIB']=oldlib
		self.env['LIBPATH']=oldlibpath
		return val
class framework_configurator(configurator_base):
	__metaclass__=attached_conf
	def __init__(self,conf):
		configurator_base.__init__(self,conf)
		self.name=''
		self.custom_code=''
		self.code='int main(){return 0;}\n'
		self.define=''
		self.path=[]
		self.uselib=''
		self.uselib_store=''
		self.remove_dot_h=False
	def error(self):
		errmsg='framework %s cannot be found via compiler, try pass -F'%self.name
		if self.message:errmsg+='\n%s'%self.message
		self.conf.fatal(errmsg)
	def validate(self):
		if not self.uselib_store:
			self.uselib_store=self.name.upper()
		if not self.define:
			self.define=self.conf.have_define(self.uselib_store)
		if not self.code:
			self.code="#include <%s>\nint main(){return 0;}\n"
	def run_test(self):
		code=[]
		if self.remove_dot_h:
			code.append('#include <%s/%s>\n'%(self.name,self.name))
		else:
			code.append('#include <%s/%s.h>\n'%(self.name,self.name))
		code.append('int main(){%s\nreturn 0;}\n'%self.custom_code)
		for p in self.path:
			linkflags=['-framework',self.name,'-F%s'%p]
			ccflags=['-F%s'%p]
			obj=check_data()
			obj.code="\n".join(code)
			obj.uselib=self.uselib_store+" "+self.uselib
			obj.env=self.env.copy()
			obj.env['LINKFLAGS']=linkflags
			obj.env['CCFLAGS']=ccflags
			obj.env['CXXFLAGS']=ccflags
			ret=int(self.conf.run_check(obj))
			if ret:break
		self.conf.check_message('framework %s'%self.name,'',ret,option='')
		self.conf.define_cond(self.define,ret)
		val={}
		if ret:
			val['LINKFLAGS_'+self.uselib_store]=linkflags
			val['CCFLAGS_'+self.uselib_store]=ccflags
			val['CXXFLAGS_'+self.uselib_store]=ccflags
			val[self.define]=ret
		self.update_env(val)
		return val
class header_configurator(configurator_base):
	__metaclass__=attached_conf
	def __init__(self,conf):
		configurator_base.__init__(self,conf)
		self.name=''
		self.path=[]
		self.header_code=''
		self.custom_code=''
		self.code='int main() {return 0;}\n'
		self.define=''
		self.nosystem=0
		self.libs=[]
		self.lib_paths=[]
		self.uselib=''
		self.uselib_store=''
	def error(self):
		errmsg='header %s cannot be found via compiler'%self.name
		if self.message:errmsg+='\n%s'%self.message
		self.conf.fatal(errmsg)
	def validate(self):
		if not self.define:
			if self.name:self.define=self.conf.have_define(self.name)
			elif self.uselib_store:self.define=self.conf.have_define(self.uselib_store)
		if not self.code:
			self.code="#include <%s>\nint main(){return 0;}\n"
		if not self.define:
			self.conf.fatal('no define given')
	def run_test(self):
		ret={}
		if self.uselib_store:
			test=self.conf.create_header_enumerator()
			test.nosystem=self.nosystem
			test.name=self.name
			test.want_message=0
			test.path=self.path
			test.env=self.env
			ret=test.run()
			if ret:
				self.env['CPPPATH_'+self.uselib_store]=ret
		code=[]
		code.append(self.header_code)
		code.append('\n')
		code.append('#include <%s>\n'%self.name)
		code.append('int main(){%s\nreturn 0;}\n'%self.custom_code)
		obj=check_data()
		obj.code="\n".join(code)
		obj.includes=self.path
		obj.uselib=self.uselib_store+" "+self.uselib
		obj.env=self.conf.env.copy()
		obj.env['LIB']=Utils.to_list(self.libs)
		obj.env['LIBPATH']=Utils.to_list(self.lib_paths)
		ret=int(self.conf.run_check(obj))
		self.conf.check_message('header %s'%self.name,'',ret,option='')
		self.conf.define_cond(self.define,ret)
		val={}
		if ret:
			val['CPPPATH_'+self.uselib_store]=self.env['CPPPATH_'+self.uselib_store]
			val[self.define]=ret
		if not ret:return{}
		return val
class common_include_configurator(header_enumerator):
	__metaclass__=attached_conf
	def run_test(self):
		header_dir=header_enumerator.run_test(self)
		if header_dir:
			header_path=os.path.join(header_dir,self.name)
			self.env.append_unique(COMMON_INCLUDES,header_path)
		return header_dir
class check_data(object):
	def __init__(self):
		self.env=''
		self.code=''
		self.flags=''
		self.uselib=''
		self.includes=''
		self.function_name=''
		self.lib=[]
		self.libpath=[]
		self.define=''
		self.header_name=''
		self.execute=0
		self.options=''
		self.force_compiler=None
		self.build_type='program'
setattr(Configure,'check_data',check_data)
def define(self,define,value,quote=1):
	assert define and isinstance(define,str)
	tbl=self.env[DEFINES]or Utils.ordered_dict()
	if isinstance(value,str):
		if quote==1:
			tbl[define]='"%s"'%str(value)
		else:
			tbl[define]=value
	elif isinstance(value,int):
		tbl[define]=value
	else:
		raise TypeError
	self.env[DEFINES]=tbl
	self.env[define]=value
def undefine(self,define):
	assert define and isinstance(define,str)
	tbl=self.env[DEFINES]or Utils.ordered_dict()
	value=UNDEFINED
	tbl[define]=value
	self.env[DEFINES]=tbl
	self.env[define]=value
def define_cond(self,name,value):
	if value:
		self.define(name,1)
	else:
		self.undefine(name)
def is_defined(self,key):
	defines=self.env[DEFINES]
	if not defines:
		return False
	try:
		value=defines[key]
	except KeyError:
		return False
	else:
		return value!=UNDEFINED
def get_define(self,define):
	try:return self.env[DEFINES][define]
	except KeyError:return None
def have_define(self,name):
	return"HAVE_%s"%Utils.quote_define_name(name)
def write_config_header(self,configfile='',env=''):
	if not configfile:configfile=WAF_CONFIG_H
	lst=Utils.split_path(configfile)
	base=lst[:-1]
	if not env:env=self.env
	base=[self.blddir,env.variant()]+base
	dir=os.path.join(*base)
	if not os.path.exists(dir):
		os.makedirs(dir)
	dir=os.path.join(dir,lst[-1])
	self.env.append_value('waf_config_files',os.path.abspath(dir))
	waf_guard='_%s_WAF'%Utils.quote_define_name(configfile)
	dest=open(dir,'w')
	dest.write('/* Configuration header created by Waf - do not edit */\n')
	dest.write('#ifndef %s\n#define %s\n\n'%(waf_guard,waf_guard))
	if not configfile in self.env['dep_files']:
		self.env['dep_files']+=[configfile]
	tbl=env[DEFINES]or Utils.ordered_dict()
	for key in tbl.allkeys:
		value=tbl[key]
		if value is None:
			dest.write('#define %s\n'%key)
		elif value is UNDEFINED:
			dest.write('/* #undef %s */\n'%key)
		else:
			dest.write('#define %s %s\n'%(key,value))
	for include_file in self.env[COMMON_INCLUDES]:
		dest.write('\n#include "%s"'%include_file)
	dest.write('\n#endif /* %s */\n'%waf_guard)
	dest.close()
def run_check(self,obj):
	if not obj.code:
		raise Configure.ConfigurationError('run_check: no code to process in check')
	dir=os.path.join(self.blddir,'.wscript-trybuild')
	for(root,dirs,filenames)in os.walk(dir):
		for f in list(filenames):
			os.remove(os.path.join(root,f))
	bdir=os.path.join(dir,'testbuild')
	if(self.env['CXX_NAME']and not obj.force_compiler and Task.TaskBase.classes.get('cxx',None))or obj.force_compiler=="cxx":
		tp='cxx'
		test_f_name='test.cpp'
	else:
		tp='cc'
		test_f_name='test.c'
	if not os.path.exists(dir):
		os.makedirs(dir)
	if not os.path.exists(bdir):
		os.makedirs(bdir)
	if obj.env:env=obj.env
	else:env=self.env.copy()
	dest=open(os.path.join(dir,test_f_name),'w')
	dest.write(obj.code)
	dest.close()
	back=os.path.abspath('.')
	bld=Build.BuildContext()
	bld.log=self.log
	bld.all_envs.update(self.all_envs)
	bld.all_envs['default']=env
	bld._variants=bld.all_envs.keys()
	bld.load_dirs(dir,bdir)
	os.chdir(dir)
	bld.rescan(bld.srcnode)
	o=bld.new_task_gen(tp,obj.build_type)
	o.source=test_f_name
	o.target='testprog'
	o.uselib=obj.uselib
	o.includes=obj.includes
	self.log.write("==>\n%s\n<==\n"%obj.code)
	try:
		ret=bld.compile()
	except Build.BuildError:
		ret=1
	if obj.execute:
		lastprog=o.link_task.outputs[0].abspath(o.env)
	os.chdir(back)
	if obj.execute:
		if ret:return not ret
		data=Utils.cmd_output('"%s"'%lastprog).strip()
		ret={'result':data}
		return ret
	return not ret
def cc_check_features(self,kind='cc'):
	v=self.env
	test=Configure.check_data()
	test.code='int main() {return 0;}\n'
	test.env=v
	test.execute=1
	test.force_compiler=kind
	ret=self.run_check(test)
	self.check_message('compiler could create','programs',not(ret is False))
	if not ret:self.fatal("no programs")
	lib_obj=Configure.check_data()
	lib_obj.code="int k = 3;\n"
	lib_obj.env=v
	lib_obj.build_type="shlib"
	lib_obj.force_compiler=kind
	ret=self.run_check(lib_obj)
	self.check_message('compiler could create','shared libs',not(ret is False))
	if not ret:self.fatal("no shared libs")
	lib_obj=Configure.check_data()
	lib_obj.code="int k = 3;\n"
	lib_obj.env=v
	lib_obj.build_type="staticlib"
	lib_obj.force_compiler=kind
	ret=self.run_check(lib_obj)
	self.check_message('compiler could create','static libs',not(ret is False))
	if not ret:self.fatal("no static libs")
def cxx_check_features(self):
	return cc_check_features(self,kind='cpp')
def check_pkg(self,modname,destvar='',vnum='',pkgpath='',pkgbin='',pkgvars=[],pkgdefs={},mandatory=False):
	pkgconf=self.create_pkgconfig_configurator()
	if not destvar:destvar=modname.upper()
	pkgconf.uselib_store=destvar
	pkgconf.name=modname
	pkgconf.version=vnum
	if pkgpath:pkgconf.pkgpath=pkgpath
	pkgconf.binary=pkgbin
	pkgconf.variables=pkgvars
	pkgconf.defines=pkgdefs
	pkgconf.mandatory=mandatory
	return pkgconf.run()
def pkgconfig_fetch_variable(self,pkgname,variable,pkgpath='',pkgbin='',pkgversion=0):
	if not pkgbin:pkgbin='pkg-config'
	if pkgpath:pkgpath='PKG_CONFIG_PATH=$PKG_CONFIG_PATH:'+pkgpath
	pkgcom='%s %s'%(pkgpath,pkgbin)
	if pkgversion:
		ret=os.popen("%s --atleast-version=%s %s"%(pkgcom,pkgversion,pkgname)).close()
		self.conf.check_message('package %s >= %s'%(pkgname,pkgversion),'',not ret)
		if ret:
			return''
	else:
		ret=os.popen("%s %s"%(pkgcom,pkgname)).close()
		self.check_message('package %s '%(pkgname),'',not ret)
		if ret:
			return''
	return Utils.cmd_output('%s --variable=%s %s'%(pkgcom,variable,pkgname)).strip()

conf(define)
conf(undefine)
conf(define_cond)
conf(is_defined)
conf(get_define)
conf(have_define)
conf(write_config_header)
conf(run_check)
conftest(cc_check_features)
conftest(cxx_check_features)
conf(check_pkg)
conf(pkgconfig_fetch_variable)
